// License declaration at end of file

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sys/stat.h>

#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/utilfuncs.h"
#include "channelflow/turbstats.h"

using namespace std;
using namespace channelflow;

int main(int argc, char* argv[]) {

  string purpose("compute statistics of a periodic orbit");

  ArgList args(argc, argv, purpose);

  const Real T        = args.getreal("-T",   "--T", 0.0, "period of orbit");
  /***/ Real N        = args.getreal("-N",   "--N", 0,   "check stats at interval dT=T/N (default dT approx 1)");
  const bool vardt    = args.getflag("-vdt", "--variabledt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg    = args.getreal("-dt",     "--dt", 0.03125, "timestep");
  const Real dtmin    = args.getreal("-dtmin",  "--dtmin", 0.001, "minimum time step");
  const Real dtmax    = args.getreal("-dtmax",  "--dtmax", dtarg,  "maximum time step");
  const Real CFLmin   = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax   = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");
  const string stepstr= args.getstr ("-ts", "--timestepping", "sbdf3", "timestepping algorithm, one of "
				     "[cnfe1 cnab2 smrk2 sbdf1 sbdf2 sbdf3 sbdf4]");
  const string nonlstr= args.getstr ("-nl", "--nonlinearity", "rot", "method of calculating nonlinearity, "
				     "one of [rot conv div skew alt]");
  const string symmstr= args.getstr ("-symms","--symmetries", "", "constrain u(t) to invariant symmetric subspace, argument is the filename for a file listing the generators of the isotropy group");
  const string odir   = args.getpath("-o", "--outdir", "orbstats/", "output directory");
  const Real Reynolds = args.getreal("-R", "--Reynolds", 400, "Reynolds number");
  const bool channel  = args.getflag("-c", "--channel", "channelflow instead of plane Couette");
  const bool bulkvel  = args.getflag("-b", "--bulkvelocity","hold bulk velocity fixed, not pressure gradient");
  const Real dPdx     = args.getreal("-P", "--dPdx",   0.0, "value for fixed pressure gradient");
  const Real Ubulk    = args.getreal("-U", "--Ubulk",  0.0, "value for fixed bulk velocity");
  const string uname  = args.getstr (1, "<flowfield>", "initial condition");
  const char s = ' ';

  args.check();
  args.save("./");
  mkdir(odir);

  cout << "Constructing u,q, and optimizing FFTW..." << endl;
  FlowField u(uname);
  u.optimizeFFTW(FFTW_PATIENT);

  const int Nx = u.Nx();
  const int Ny = u.Ny();
  const int Nz = u.Nz();
  const Real Lx = u.Lx();
  const Real Lz = u.Lz();
  const Real a = u.a();
  const Real b = u.b();

  FlowField q(Nx,Ny,Nz,1,Lx,Lz,a,b);

  if (N==0)
    N = iround(T);
  const Real dT = T/N;
  TimeStep dt(dtarg, dtmin, dtmax, dT, CFLmin, CFLmax, vardt);

  cout << " T == " << T << endl;
  cout << " N == " << N << endl;
  cout << "dT == " << dT<< endl;
  cout << "dt == " << dt<< endl;

  const Real nu = 1.0/Reynolds;

  // Construct time-stepping algorithm
  DNSFlags flags;
  flags.timestepping = s2stepmethod(stepstr);
  flags.dealiasing   = DealiasXZ;
  flags.nonlinearity = s2nonlmethod(nonlstr);
  flags.constraint   = bulkvel ? BulkVelocity : PressureGradient;
  flags.baseflow     = channel ? Parabolic : PlaneCouette;
  flags.dPdx         = dPdx;
  flags.Ubulk        = Ubulk;

  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    flags.symmetries = symms;
  }

  cout << "constructing DNS..." << endl;
  DNS dns(u, nu, dt, flags, 0.0);

  ChebyCoeff U(Ny,a,b,Spectral);
  if (channel) {
    U[0] = 0.5;
    U[2] = -0.5;
  }
  else
    U[1] = 1;

  TurbStats stats(U, nu);
  ofstream os((odir + "IDEt.asc").c_str());
  os << "% ";
  for (int n=0; n<argc; ++n)
    os << argv[n] << s;
  os << endl;
  os << "% wallshear(u+U) dissipation(u+U) energy(u+U) energy(u)" << endl;
  os << setprecision(17);
  Real I = 0.0;
  Real D = 0.0;
  Real E = 0.0;
  Real En = 0.0;
  int count = 0;
  os << setprecision(17);

  FlowField umean(u);
  umean.setToZero();

  FlowField tmp(Nx,Ny,Nz,9,Lx,Lz,a,b);

  // Don't close the orbit so as not to count last point twice
  for (Real t=0; t<=T-dT/2; t += dT) {

    umean += u;
    Real en = 0.5*L2Norm2(u);

    u += U;
    Real i = wallshear(u);
    Real d = dissipation(u);
    Real e = 0.5*L2Norm2(u);
    u -= U;

    os << i <<s<< d <<s<< e <<s<< en << endl;

    I += i;
    D += d;
    E += e;
    En += en;
    stats.addData(u,tmp);
    ++count;

    cout << "          t == " << t << endl;
    cout << "  L2Norm(u) == " << L2Norm(u) << endl;
    cout << "dissip(u+U) == " << d << endl;
    cout << " shear(u+U) == " << i << endl;
    cout << "energy(u+U) == " << e << endl;

    dns.advance(u, q, dt.n());

    if (dt.adjust(dns.CFL()))
      dns.reset_dt(dt);
  }

  ofstream os2((odir + "IDE.asc").c_str());
  os2 << "% <wallshear(u+U)> <dissipation(u+U)> <energy(u+U)> <energy(u)>" << endl;
  os2 << I/count <<s<< D/count <<s<< E/count <<s<< En/count<< endl;

  umean *= 1.0/count;
  umean.save(odir + "umean");

  stats.msave(odir + "stats", false);
  stats.msave(odir + "wallstats", true);
  cout << "done!" << endl;
}

/* couette.cpp: time-integration class for spectral Navier-Stokes simulator
 * channelflow-1.1
 *
 * Copyright (C) 2001-2009  John F. Gibson
 *
 * gibson@cns.physics.gatech.edu  johnfgibson@gmail.com
 *
 * Center for Nonlinear Science
 * School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 * 404 385 2509
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation version 2
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
